from sklearn.base import BaseEstimator, clone
import numpy as np
import warnings
import pandas as pd
import mlflow
import pickle
import os
import tqdm as tqdm
from xgboost import XGBRegressor, XGBClassifier



class StepShifter(BaseEstimator):
    """
    Introduction
    Stepshifter3 is a general purpose stepshifting algorithm for tabular data based on Hegre et.al 2019.
    It is designed to be used with any scikit-learn compatible estimator,
    and can be used for both regression and classification.
    It is further designed to be used with tabular data, and can handle both single and multi-index DataFrames.

    Definitions
    - The VIEWS month id is a counter that starter on 1. January 1980


    Variables
    τ           = reference time for which the project is operating,
                   we can observe values less than tau but not greater
    S           = maximum number of steps ahead to predict
    M           = the set of months in a test window, i.e., the months we want to iterate stepshifts over
    s           = step number, i.e., the number of steps ahead we want to predict
    e           = superscript denoting evaluation time partition
    τ^e_0       = training partition period starting point
    τ^e_t       = training partition period ending point
    τ_0         = forceast period starting point
    τ_t         = forecast period ending point
    τ^e_t + 1   = evaluation calibration period starting point
    τ^e_c       =  evaluation calibration period ending point
    τ^e^c + 1   = evaluation forecast period starting point
    τ^e_f       =  forecast period ending point
    τ_t+1       = forecast period starting point
    τ_c         = forecast period ending point
    m_j         = model specification

    Example:
    ___________________________________________________________________________________________________
    |                         | Periodization                                                          |
    |--------------------------------------------------------------------------------------------------|
    |                         | Evaluation                       |  Forecast                           |
    |-------------------------|----------------------------------|-------------------------------------|
    | Training Period         | τ^e_0     = 121 (January 1990)   | τ_0          = 121 (January 1990)   |
    |_________________________| τ^e_t     = 396 (December 2012)  | τ_t          = 432 (December 2015)  |
    | Calibration Period      | τ^e_t + 1 = 397 (January 2013)   | τ_t + 1      = 433 (January 2016)   |
    |_________________________| τ^e_c     = 432 (December 2015)  | τ_c          = 468 (December 2018)  |
    | Predictor Updating      | n/a                              | τ_c + 1      = 469 (January 2019)   |
    |_________________________| n/a                              | τ_c + (k −1) = 480 (December 2019)  |
    | Testing Period          | τ^e_c + 1 = 433 (January 2016)   | τ_c + k      = 481 (January 2020)   |
    |                         | τ^e_f     = 468 (December 2018)  | τ_f          = 516 (December_2022)  |
    |--------------------------------------------------------------------------------------------------|

    Pseudocode for stepshifter3:

    1.
        for each model m_j in M:
            for each step s in S:
                model m_j is trained on the training partition period tau_0^e to tau_t^e
    2.
        generate predictions for each m^(j,s)
        for all months i in the calibration period tau_t^e + 1 to tau_c^e
        using data up to s months before i

    3.
        Calibrate models, obtain ensamble weights for each model,
        and tune hyperparameters using the predictions from step 2;
        and the actuals for all months i in the calibration period tau_t^e + 1 to tau_c^e

    4.
        Retrain model m^(j,s) using both the training partition period,
        tau_0^e to tau_t^e and the calibration period tau_t^e + 1 to tau_c^e

    5.
        Generate predictions for the testing/forecasting period tau_^c_e + 1 to tau_^c_f

    Args:
        base_estimator (object): The estimator to be used for training and prediction.
        df (DataFrame): The DataFrame to be used for training and prediction.
        config (dict): A dictionary containing the configuration for the model,
        see the example config file for more information.

    Authors: Tom Daniel Grande, Jonas Schie Olsen
    Written: 2023-08-28 -> ????-??-??
"""

    def __init__(self, base_estimator,  df=None, config=None):
        self.base_estimator = base_estimator  # The wrapped estimator, could be xgboost, sklearn, etc.
        self.S = None  # Maximum steps ahead to predict
        self.t = None  # Training period starting point
        self.t_1 = None  # Training period ending point
        self.M_ssa = None  # The stepcombined model object
        self.models = {}  # Dictionary to store trained models for each step
        self.predictions = None
        self.metrics_report = None  # If True, will generate a metrics report
        self.run_name = None  # Name of the run
        self.ID_columns = None
        self.time_column = None
        self.t_0 = None
        self.X = None
        self.y = None
        self.mlflow_tracking_uri = None
        self.experiment_name = None
        self.is_xgb = None
        self.combine_method = None
        self._set_internal_variables_from_config(config, df)

    def _set_internal_variables_from_config(self, config, df):
        """Sets internal variables from a config file."""
        self.S = config['S']
        self.t_0 = config['tau^e_0']
        self.t_1 = config['tau^e_t']
        self.ID_columns = config['ID_columns']
        self.mlflow_tracking_uri = config['mlflow_tracking_uri']
        self.metrics_report = config['metrics_report']
        self.experiment_name = config['experiment_name']
        self.run_name = config['run_name']
        self.df = self._validate_df(df, config)
        self.X = self._validate_X(config)
        self.y = self._validate_y(config)
        self.time_column = config['time_column']
        self.is_xgb = (isinstance(self.base_estimator, XGBRegressor) or isinstance(self.base_estimator, XGBClassifier))
        self.combine_method = config['combine_method']
        # Checks on parameters:

        # Check if S is an integer
        if not isinstance(self.S, int):
            raise ValueError("max_steps must be an integer.")
        if self.S > 1024:
            warnings.warn("You have specified more than 1024 steps, this could impact performance.")

    def _validate_df(self, df, config):
        """
        Remove all countries that do not have rows in the range (t1, t2).

        Parameters:
            df (DataFrame): The original DataFrame.
            t1 (int): The start of the time range.
            t2 (int): The end of the time range.

        Returns:
            DataFrame: A filtered DataFrame.
        """
        # Get unique country IDs
        unique_countries = df.index.get_level_values(1).unique()

        # List to hold the countries that meet the criteria
        valid_countries = []

        for country_id in unique_countries:
            # Filter the DataFrame for the specific country
            df_country = df.loc[pd.IndexSlice[:, country_id], :]

            # Get unique months for this country
            country_months = df_country.index.get_level_values(0).unique()

            # Check if the country has data for the entire range
            if all(month in country_months for month in range(config['tau^e_0'], config['tau^e_f'])):
                valid_countries.append(country_id)

        # Filter the DataFrame to only include valid countries
        df_filtered = df.loc[pd.IndexSlice[:, valid_countries], :]

        # throw a warning if the number of rows is reduced:
        if len(df_filtered) < len(df):
            warnings.warn("The number of rows has been reduced due to missing values.")
            warnings.warn(f"Number of rows before filtering: {len(df)}, after filtering: {len(df_filtered)}")

        return df_filtered

    def _validate_params(self):
        """Validates the parameters."""
        if not isinstance(self.S, int):
            raise ValueError("max_steps must be an integer.")
        if self.S > 1024:
            warnings.warn("You have specified more than 1024 steps, this could impact performance.")

    def _validate_X(self, config):
        """Validates the X input."""
        if not isinstance(self.df, pd.DataFrame):
            raise ValueError("df must be a pandas DataFrame.")

        # X is everything but the column config["target _column"]:
        X_cols = [col for col in self.df.columns if col not in config["target_column"]]
        X = self.df[X_cols]

        if X.shape[1] < 1:
            raise ValueError("X must have at least 3 columns (2 ID columns and 1 feature column).")
        # remove ID columns and store them for perservance in the predict

        return X

    def _validate_y(self, config):
        """Validates the y input."""
        # y is the last column in the dataframe:
        y = self.df[config["target_column"]]

        # check if y is numeric
        if not np.issubdtype(y.dtype, np.number):
            raise ValueError("Target (y) must be numeric.")

        # check if y is nan:
        if y.isna().any():
            raise ValueError("Target (y) must not contain NaNs.")
        return y

    def start_mlflow(self):
        """Starts an mlflow run."""

        mlflow.set_tracking_uri(self.mlflow_tracking_uri)

    def fit_mlflow(self, tau_start, tau_end):

        mlflow.set_experiment(self.experiment_name)
        with mlflow.start_run(run_name=self.run_name):
            mlflow.log_param("base_estimator_summary", str(self.base_estimator)[:250])  # Truncated summary
            mlflow.log_param("max_steps", self.S)

            self.models = {}  # Clear previous models
            for m in tqdm.tqdm(range(tau_start, tau_end)):  # Loop over each month
                # Initialize a dictionary to store models for this month
                self.models[m] = {}
                X_m = self.X.loc[slice(m, (tau_end)), slice(None), :]  # Filter data for the range
                y_m = self.y.loc[X_m.index]  # Filter labels accordingly

                S_max = min(self.S, tau_end-m)
                for s in range(1, S_max+1):  # Loop over each step
                    # dynamic loop over s:
                    with mlflow.start_run(run_name=f"s_{s}_m_{m}", nested=True):  # Nested run for each step and month
                        second_slice = tau_end - s

                        # Filter data to include rows from t_0 to current month
                        X_m_s = self.X.loc[slice(m, second_slice), slice(None), :]  # Filter data for the range

                        # Shift y values according to the step
                        y_shifted = y_m.groupby(level=1).shift(-s)

                        y_m_s = y_shifted.loc[X_m_s.index]
                        # Drop rows where y has been shifted to NaN

                        # check that X_m_s and y_m_s have the same length after shift
                        assert len(X_m_s) == len(y_m_s)

                        if not np.issubdtype(y_m_s.dtype, np.number):
                            warnings.warn("Target (y) must be numeric.")
                        # check that y_m_s does not contain NaNs or inf:
                        if y_m_s.isna().any():
                            # locate the nan positions:
                            # count the number of nan values:
                            nan_count = y_m_s.isna().sum()
                            nan_positions = self._debug_find_nan_positions(y_m_s)

                            warnings.warn(f"y contain NaNs at positions:{nan_positions}, the number of NaNs is: {nan_count}")
                            # replace the NaNs with 0:
                            y_m_s = y_m_s.fillna(0)
                        if y_m_s.isin([np.inf, -np.inf]).any():
                            warnings.warn("Target (y) must not contain inf.")
                        # Remove ID columns
                        # X_m_s = self.remove_ID_columns(X_m_s)

                        # Validate if y_s contains both ints and floats, if required
                        if not np.issubdtype(y_shifted.dtype, np.number):
                            warnings.warn("Target (y) must be numeric.")

                        model = clone(self.base_estimator)
                        if self.is_xgb:
                            model.fit(X_m_s, y_m_s, verbose=0, eval_set=[(X_m_s, y_m_s)])
                        else:
                            model.fit(X_m_s, y_m_s, eval_set=[(X_m_s, y_m_s)])

                        # Calculate metrics
                        y_hat = model.predict(X_m_s)
                        rmse = np.sqrt(np.mean((y_m_s - y_hat)**2))
                        mlflow.log_metric("rmse", rmse)

                        # Calculate accuracy
                        accuracy = np.mean(y_m_s == y_hat)
                        mlflow.log_metric("accuracy", accuracy)

                        # Store the model
                        self.models[m][s] = model
                        mlflow.sklearn.log_model(model, f"model_step_{s}_month_{m}")

                        # Pickle the model and log it as an artifact
                        with open(f"model_step_{s}_month_{m}.pkl", "wb") as f:
                            pickle.dump(model, f)
                        mlflow.log_artifact(f"model_step_{s}_month_{m}.pkl")

                        # Remove the pickled model locally
                        os.remove(f"model_step_{s}_month_{m}.pkl")
            return self.models

    def fit(self, tau_start, tau_end):
        """Train models on shifted steps for each month."""

        self.models = {}  # Clear previous models

        for m in tqdm.tqdm(range(tau_start, tau_end)):  # Loop over each month
            # Initialize a dictionary to store models for this month
            self.models[m] = {}
            X_m = self.X.loc[slice(m, (tau_end)), slice(None), :]  # Filter data for the range
            y_m = self.y.loc[X_m.index]  # Filter labels accordingly

            S_max = min(self.S, tau_end-m)
            for s in range(1, S_max + 1):  # Loop over each step
                # dynamic loop over s:
                second_slice = tau_end-s

                # Filter data to include rows from t_0 to current month
                X_m_s = self.X.loc[slice(m, second_slice), slice(None), :]  # Filter data for the range

                # Shift y values according to the step
                y_shifted = y_m.groupby(level=1).shift(-s)

                y_m_s = y_shifted.loc[X_m_s.index]
                # check that X_m_s and y_m_s have the same length after shift
                assert len(X_m_s) == len(y_m_s)

                if not np.issubdtype(y_m_s.dtype, np.number):
                    warnings.warn("Target (y) must be numeric.")
                # check that y_m_s does not contain NaNs or inf:
                if y_m_s.isna().any():
                    # locate the nan positions:
                    # count the number of nan values:
                    nan_count = y_m_s.isna().sum()
                    nan_positions = self._debug_find_nan_positions(y_m_s)

                    warnings.warn(f"y_m_s contains NaNs at positions: {nan_positions}, the number of NaNs is: {nan_count}")
                    # replace the NaNs with 0:
                    y_m_s = y_m_s.fillna(0)
                if y_m_s.isin([np.inf, -np.inf]).any():
                    warnings.warn("Target (y) must not contain inf.")
                # Remove ID columns
                # X_m_s = self.remove_ID_columns(X_m_s)

                # Validate if y_s contains both ints and floats, if required
                if not np.issubdtype(y_shifted.dtype, np.number):
                    warnings.warn("Target (y) must be numeric.")

                # Train the estimator for this step

                model = clone(self.base_estimator)
                if self.is_xgb:
                    model.fit(X_m_s, y_m_s, verbose=0, eval_set=[(X_m_s, y_m_s)])
                else:
                    model.fit(X_m_s, y_m_s, eval_set=[(X_m_s, y_m_s)])

                # Store the model for this month and step:
                self.models[m][s] = model
        return self.models

    def predict(self, tau_start, tau_end):
        """Generate predictions for the testing period."""
        prediction_matrix = []
        for taus in range(tau_start, tau_end):
            X_m = self.X.loc[slice(taus, taus), slice(None), :]  # Do prediction for one month at a time
            predictions_for_month = []

            # current max step we have a true value for:
            S_max = min(self.S, tau_end-taus)

            # loop through all models which we have trained on
            for m in range(list(self.models.keys())[0], list(self.models.keys())[-1]+1):
                predictions_for_step = []
                # loop through all steps for each month, but exlude steps we don't have a true value for
                for s in range(1, min(len(self.models[m]), S_max)):
                    model = self.models[m][s]
                    prediction = model.predict(X_m)
                    predictions_for_step.append(prediction)
                predictions_for_month.append(predictions_for_step)
            prediction_matrix.append(predictions_for_month)
        self.predictions = prediction_matrix
        return prediction_matrix

    def predict_ansamble(self, X_test, steps):
        """ Predicts the ansamble of models for each step.
        Input:
            X_test (pd.DataFrame): the test set.
            steps (list): list of steps to predict. i.e. [1, 2, 3, 4, 10, 36](Optional)
        Returns:
            predictions (list): list of predictions for each step. shape: [len(steps), len(X_test)]
        """
        # check that passed X only has one month:
        assert len(X_test.index.get_level_values(0).unique()) == 1, "X_test must only contain one month"

        if self.M_ssa['method'] == ('mean_of_all' or 'mean_of_best'):
            predictions = []
            for s in range(1, len(self.M_ssa['models'])+1):
                preds_per_model = []
                for model in self.M_ssa['models'][s-1]:
                    preds_per_model.append(model.predict(X_test))

                predictions.append(np.mean(preds_per_model, axis=0))

        else:
            pass
            # TODO implement other combine methods later on demand

        if steps is None:
            print("returning all steps")
            steps = list(range(1, self.S+1))
        else:
            pass

        # return only the predictions for the steps specified in steps
        return [predictions[i-1] for i in steps]

    def predict_mlflow(self, X, logged_model):
        # TODO: implement this
        pass

    def stepcombine(self):
        """Combine the relevant models for eavh step based on chosen combine method.

        Returns:
            M_ssa (dict): A dictionary containing the combined models in a list of lists,
              where each list is the models associated with a step s and the combine method used.
        """
        # TODO: implement more combine methods
        if self.combine_method == 'mean_of_all':
            self.M_ssa = {
                'models': self.mean_of_all_models(),
                'method': 'mean_of_all',
                'description': 'The mean of all models from training period [tau^e_0, tau^e_t], '
                'for each step is used as the prediction for that step'

            }
        elif self.combine_method == 'mean_of_best':
            self.M_ssa = {
                'models': self.mean_of_best_models(),
                'method': 'mean_of_best',
                'description': 'The mean of the best model from training period [tau^e_0, tau^e_t],'
                'for each step is used as the prediction for that step'
            }
        else:
            raise ValueError('The combine method must be one of the following: [mean_of_all,mean_of_best] ')

        return self.M_ssa

    def mean_of_all_models(self):
        # loop through all steps and add the models per step to a list.
        list_of_all_models = []
        for s in range(1, self.S+1):
            predictions_for_step = []
            # loop through all steps for each month, but exlude steps we don't have a true value for
            for m in range(list(self.models.keys())[0], list(self.models.keys())[-1]+1):
                # append the model if it exist:
                if self.models[m].get(s, None) is not None:
                    predictions_for_step.append(self.models[m][s])
                else:
                    pass
            list_of_all_models.append(predictions_for_step)
        return list_of_all_models

    def mean_of_best_models(self):
        # loop through all steps and add the models per step to a list.
        pass

    def get_step_number(self, step_number):
        """Retrieve the model for a given step number."""
        return self.models.get(step_number)

    def remove_ID_columns(self, df) -> pd.DataFrame:
        if isinstance(df.index, pd.MultiIndex):
            # Reset the index if it's a multi-index DataFrame
            df_reset = df.reset_index()

            # Only drop the columns specified in self.id_columns if they exist
            drop_columns = [col for col in self.ID_columns if col in df_reset.columns]
            df_clean = df_reset.drop(drop_columns, axis=1)

        else:
            # For single-index DataFrames, drop the columns specified in self.id_columns if they exist
            drop_columns = [col for col in self.ID_columns if col in df.columns]
            df_clean = df.drop(drop_columns, axis=1)

        return df_clean

    def generate_metrics_report(self):
        """Placeholder for generating a metrics report."""
        # To be implemented, will probably use the GenerateReport class:
        # GenerateReport(model, X, y)

    """
    DEBUG FUNCTIONS:
    """

    @staticmethod
    def _debug_visualize_step(y_1, y_2):
        # filter y to only include one country:
        y_1 = y_1.loc[y_1.index.get_level_values(1) == 3]
        y_2 = y_2.loc[y_2.index.get_level_values(1) == 3]
        pd.set_option('display.max_rows', None)
        # add both to same dataframe:
        y_1 = y_1.to_frame()
        y_2 = y_2.to_frame()
        y_1.columns = ['y_1']
        y_2.columns = ['y_2']
        y_1['y_2'] = y_2['y_2']
        print(y_1)

    @staticmethod
    def _debug_find_nan_positions(y):
        nan_positions = np.where(y.isna())[0]
        # Return also the country id for the nan positions:
        nan_positions = y.index.get_level_values(1)[nan_positions]
        return nan_positions

    @staticmethod
    def _debug_old_erronoues_predict(self, tau_start, tau_end):
        # Initialize the list of lists to store the predictions
        prediction_matrix = []

        for m in range(tau_start, tau_end):
            m_list = []
            for s in range(1, self.S + 1):  # Step horizon starts from 1 to S
                s_list = []
                second_slice = m + s
                X_m_s = self.X.loc[slice(m, second_slice), slice(None), :]

                for past_month in range(m - self.S, m):
                    step_for_past_month = m - past_month - 1
                    model = self.models.get(past_month, {}).get(step_for_past_month, None)

                    if model is not None:
                        prediction = model.predict(X_m_s)
                        s_list.append(prediction)
                    else:
                        pass

                m_list.append(s_list)

            prediction_matrix.append(m_list)
